!========================================================================
! File: user.f90.default
!
! This is a template for user-defined functions to be called by WHIZARD.
! Copy this code to 'user.f90' and modify it according to your needs.
! The code will automatically be compiled and included in the WHIZARD
! executable. 
!
! There are four places where user code can be inserted:
! 1) User-defined beam spectra
! 2) User-defined structure functions
! 3) User-defined cut algorithms
! 4) User-defined reweighting functions for the squared matrix element
! 5) User-defined fragmentation
!========================================================================

module user

  use kinds, only: default !NODEP!
  use kinds, only: double !NODEP!
  use lorentz
  use particles

  implicit none
  private

  public :: spectrum_single, spectrum_double
  public :: spectrum_prt_out, spectrum_beam_remnant

  public :: strfun_single, strfun_double
  public :: strfun_prt_out, strfun_beam_remnant

  public :: pdf_init, pdf_strfun_array

  public :: cut
  public :: weight

  public :: fragment_init, fragment_call

! module variables used by the structure function part
  integer :: pdf_code_in, pdf_ngroup, pdf_nset, pdf_nfl, pdf_lo
  real(kind=double) :: pdf_QCDL4, pdf_top_mass

contains

!========================================================================

! While the physical meaning of a 'spectrum' and a 'structure function' 
! is different, the implementation is exactly analogous.
! The difference is that a user spectrum will be applied BEFORE all built-in
! spectra and structure functions (if any), while a user structure functions
! will be applied AFTER all built-ins.
!
! Four functions have to be coded by the user:
!   spectrum_single, spectrum_double, spectrum_prt_out, spectrum_beam_remnant
! Except for 'spectrum_single', the templates below can be left unchanged,
! if the default behavior is appropriate.
!
! The user spectrum functions take two parameters which can be set in
! the WHIZARD input file:
!   sqrts  : default real (if not set, the total c.m. energy of the process)
!   mode   : integer      (to distinguish among different user spectra)
! In the input file, these are called USER_spectrum_sqrts and 
! USER_spectrum_mode, resp. USER_strfun_sqrts and USER_strfun_mode

!------------------------------------------------------------------------
! Return the outgoing particle PDG code, given the incoming particle code
! and the mode.  Note that named constants for PDG codes are defined in the
! 'particles.f90' module and are available here.
! For a beam spectrum, this is usually trivial.

  function spectrum_prt_out (prt_in, mode) result (prt_out)
    integer, intent(in) :: prt_in, mode
    integer :: prt_out
    select case (mode)
    case default
       prt_out = prt_in
    end select
  end function spectrum_prt_out

!------------------------------------------------------------------------
! Return the PDG code of the beam remnant, given the incoming particle code
! and the mode.
! For a beam spectrum, normally there is no beam remnant, and we return
! UNDEFINED (=0)

  function spectrum_beam_remnant (prt_in, mode) result (prt_out)
    integer, intent(in) :: prt_in, mode
    integer :: prt_out
    select case (mode)
    case default
       prt_out = UNDEFINED
    end select
  end function spectrum_beam_remnant
    
!-----------------------------------------------------------------------
! The correlated spectrum for a beam pair.
! Given the two input parameters x(1), x(2), which are uniformly distributed
! between 0 and 1, the spectrum function returns output x values together
! with 'factor', which is the spectrum times the Jacobian of the x mapping,
! if any.
!
! The numbers returned may also depend on the helicities of the incoming
! and outgoing particles.  These are specified in the form of complex density
! matrices rho_in(:,:,i) and rho_out(:,:,i), where i is the beam index.  
!
! Each density matrix is a two-dimensional array, where the diagonal elements 
! should be real, and the trace should be normalized to unity.  For states
! with definite helicity, only the diagonal elements are set.  Off-diagonal
! elements refer to superpositions of helicity states, e.g., transversal
! polarization of fermions. 
! The norm of the spectrum, i.e., the spectrum averaged over initial
! helicities and summed over outgoing helicities, should be returned 
! as the overall 'factor'.
! The non-zero entries in the density matrices are defined as +-1 for fermions,
! +-1 for massless vector bosons, +-1 and 0 for massive vector bosons, and 0 
! for scalars.  Note that the correct assignment is not checked.  
!
! If the spectrum does not depend on polarization, one should set 
!   rho_out = rho_in   if the in- and out-particles are identical, resp. 
!   rho_out = unpolarized_density_matrix (prt_out)   if they are not.
! The function 'unpolarized_density_matrix' is defined in the module 
! 'particles', as are the PDG codes of the particles.
!
! For factorized spectra, the code below is appropriate.  Modify it if
! spectra are correlated, or if you want to improve the performance by 
! mapping the unit square onto itself.

  subroutine spectrum_double (factor, x, prt_in, sqrts, rho_in, rho_out, mode)
    real(kind=default), intent(out) :: factor
    real(kind=default), dimension(2), intent(inout) :: x
    integer, dimension(2), intent(in) :: prt_in
    real(kind=default), dimension(2), intent(in) :: sqrts
    complex(kind=default), dimension(-1:1,-1:1,2), intent(in) :: rho_in
    complex(kind=default), dimension(-1:1,-1:1,2), intent(out) :: rho_out
    integer, dimension(2), intent(in) :: mode
    real(kind=default), dimension(2) :: f
    integer :: i
    factor = 1
    do i = 1, 2
       call spectrum_single (f(i), x(i), prt_in(i), sqrts(i), &
            &                rho_in(:,:,i), rho_out(:,:,i), mode(i))
       factor = factor * f(i)
    end do
  end subroutine spectrum_double

!-----------------------------------------------------------------------
! The spectrum for a single beam
! Given the input parameter x, which is initially uniformly distributed
! between 0 and 1, the spectrum function returns an output x value together
! with 'factor', which is the unpolarized spectrum times the Jacobian of the
! x mapping, if any.
! The outgoing density matrix must also be set.  See the comment for
! spectrum_double.
!
! The code below describes a trivial spectrum which is unity over the
! whole kinematical range.  Modify this according to your needs.
! For correlated spectra, leave the code below as is and modify the subroutine
! spectrum_double instead.

  subroutine spectrum_single (factor, x, prt_in, sqrts, rho_in, rho_out, mode)
    real(kind=default), intent(out) :: factor
    real(kind=default), intent(inout) :: x
    integer, intent(in) :: prt_in
    real(kind=default), intent(in) :: sqrts
    complex(kind=default), dimension(-1:1,-1:1), intent(in) :: rho_in
    complex(kind=default), dimension(-1:1,-1:1), intent(out) :: rho_out
    integer, intent(in) :: mode
    integer :: prt_out
    prt_out = spectrum_prt_out (prt_in, mode)
    select case (mode)
    case default       ! Trivial spectrum without polarization dependence
       factor = 1
       if (prt_in == prt_out) then
          rho_out = rho_in
       else
          rho_out = unpolarized_density_matrix (prt_out)
       end if
    end select
  end subroutine spectrum_single

!========================================================================

! The structure function definitions are analogous to spectrum definitions
!
! Four functions have to be coded by the user:
!   strfun_single, strfun_double, strfun_prt_out, strfun_beam_remnant
! Except for 'strfun_single', the templates below can be left unchanged,
! if the default behavior is appropriate.
!
! The user functions take two parameters which can be set in
! the WHIZARD input file:
!   sqrts  : default real (if not set, the total c.m. energy of the process)
!   mode   : integer      (to distinguish among different user spectra)
! In the input file, these are called USER_strfun_sqrts and 
! USER_strfun_mode, resp. USER_strfun_sqrts and USER_strfun_mode

!------------------------------------------------------------------------
! Return the outgoing particle PDG code, given the incoming particle code
! and the mode.  Note that named constants for PDG codes are defined in the
! 'particles.f90' module and are available here.
! Default: incoming=outgoing (like, e.g., ISR)

  function strfun_prt_out (prt_in, mode) result (prt_out)
    integer, intent(in) :: prt_in, mode
    integer :: prt_out
    select case (mode)
    case default
       prt_out = prt_in
    end select
  end function strfun_prt_out

!------------------------------------------------------------------------
! Return the PDG code of the beam remnant, given the incoming particle code
! and the mode.
! Insert something sensible here if you want to have it in the event record.

  function strfun_beam_remnant (prt_in, mode) result (prt_out)
    integer, intent(in) :: prt_in, mode
    integer :: prt_out
    select case (mode)
    case default
       prt_out = UNDEFINED
    end select
  end function strfun_beam_remnant
    
!-----------------------------------------------------------------------
! The correlated structure functions for a beam pair.
! Given the two input parameters x(1), x(2), which are uniformly distributed
! between 0 and 1, the strfun function returns output x values together
! with 'factor', which is the strfun times the Jacobian of the x mapping,
! if any.
! For the helicity treatment, see the comments for spectrum_double above.
!
! Given the fact that structure functions are normally factorized, this may
! be left as is.  However, you can improve the performance by 
! mapping the unit square.

  subroutine strfun_double (factor, x, prt_in, sqrts, rho_in, rho_out, mode)
    real(kind=default), intent(out) :: factor
    real(kind=default), dimension(2), intent(inout) :: x
    integer, dimension(2), intent(in) :: prt_in
    real(kind=default), dimension(2), intent(in) :: sqrts
    complex(kind=default), dimension(-1:1,-1:1,2), intent(in) :: rho_in
    complex(kind=default), dimension(-1:1,-1:1,2), intent(out) :: rho_out
    integer, dimension(2), intent(in) :: mode
    real(kind=default), dimension(2) :: f
    integer :: i
    factor = 1
    do i = 1, 2
       call strfun_single (f(i), x(i), prt_in(i), sqrts(i), &
            &              rho_in(:,:,i), rho_out(:,:,i), mode(i))
       factor = factor * f(i)
    end do
  end subroutine strfun_double

!-----------------------------------------------------------------------
! The structure function for a single beam/parton
! Given the input parameter x, which is initially uniformly distributed
! between 0 and 1, the strfun function returns an output x value together
! with 'factor', which is the strfun times the Jacobian of the x mapping,
! if any.
!
! The code below describes a trivial structure function which is unity over the
! whole kinematical range.  Modify this according to your needs.
! For correlated spectra, leave the code below as is and modify the subroutine
! strfun_double instead.

  subroutine strfun_single (factor, x, prt_in, sqrts, rho_in, rho_out, mode)
    real(kind=default), intent(out) :: factor
    real(kind=default), intent(inout) :: x
    integer, intent(in) :: prt_in
    real(kind=default), intent(in) :: sqrts
    complex(kind=default), dimension(-1:1,-1:1), intent(in) :: rho_in
    complex(kind=default), dimension(-1:1,-1:1), intent(out) :: rho_out
    integer, intent(in) :: mode
    integer :: prt_out
    prt_out = spectrum_prt_out (prt_in, mode)
    select case (mode)
    case default      ! Trivial function without polarization dependence
       factor = 1
       if (prt_in == prt_out) then
          rho_out = rho_in
       else
          rho_out = unpolarized_density_matrix (prt_out)
       end if
    end select
  end subroutine strfun_single

!=======================================================================
! This interface allow for inserting user-defined parton distribution 
! functions or PDFs which are not (yet) contained in the standard PDFlib

! This interface will be selected if PDF_ngroup is negative in the
! WHIZARD input file, while a positive value will select the corresponding
! PDFLIB structure functions.

! Available parameters and module variables:
!   code_in     : PDG code of the beam particle (e.g., PROTON)
!   ngroup      : Absolute value of PDF_ngroup
!   nset        : Value of PDF_nset
!   nfl         : Number of active flavors (PDF_nfl)
!   lo          : QCD order (PDF_lo)
!   QCDL4       : Value of Lambda(QCD)_4 (PDF_QCDL4)
!   top_mass    : Top mass value to be used

! Initialization call:
  subroutine pdf_init (pdg_code_in, ngroup, nset, nfl, lo, QCDL4, top_mass)
    integer, intent(in) :: pdg_code_in, ngroup, nset, nfl, lo
    real(kind=double), intent(in) :: QCDL4, top_mass
    pdf_code_in  = pdg_code_in
    pdf_ngroup   = ngroup
    pdf_nset     = nset
    pdf_nfl      = nfl
    pdf_lo       = lo
    pdf_QCDL4    = QCDL4
    pdf_top_mass = top_mass
    ! More initialization here
  end subroutine pdf_init
  
! Structure function call:
! Input parameters:
!   x           : Bjorken x value
!   scale       : Q scale (set by PDF_running_scale or PDF_scale)
! Output:
!   dxpdf       : array of PDF values [x*f(x)]
!               : (tbar,bbar,cbar,sbar,ubar,dbar,g,d,u,s,c,b,t)
!               :   -6   -5   -4   -3   -2   -1  0 1 2 3 4 5 6
  subroutine pdf_strfun_array (x, scale, dxpdf)
    real(kind=double), intent(in) :: x, scale
    real(kind=double), dimension(-6:6), intent(out) :: dxpdf
    dxpdf = 0   ! Modify this as needed
  end subroutine pdf_strfun_array

!=======================================================================

! This minimal cut function accepts everything

  function cut (p, code, mode) result (accept)
    type(four_momentum), dimension(:), intent(in) :: p
    integer, dimension(:), intent(in) :: code
    integer, intent(in) :: mode
    logical :: accept
    select case (mode)
    case default
       accept = .true.
    end select
  end function cut

! ! This is an example of a cut function which simulates a 'trigger' at
! !   a hadron collider detector.  It may be adapted to your special needs,
! !   placed instead of the no-op routine above, and activated by setting
! !   user_cut_mode nonzero in the input file.  The value of user_cut_mode
! !   is transferred as the mode argument of the cut function.
! ! The routine is written such that it scans the particle codes, therefore
! !   it applies to any process.  This is recommended practice, although it
! !   is possible to write process-dependent code by explicitly adressing
! !   particle indices.
! ! In this sample cut function, the event is accepted if 
! !  (1) all partons are 0.4 units in eta-phi space apart from each other
! !  (2) There is at least
! !     - a lepton with pT > 20 or
! !     - missing pT > 50
! !     - a b quark with pT > 80 or
! !     - a jet (quark/gluon) with pT > 100 GeV
! ! Note that kinematical functions are available in the module lorentz.f90
! !   while the PDG codes E_LEPTON etc. and functions such as [[visible]]
! !   are defined in particles.f90

!   function cut (p, code, mode) result (accept)
!     ! function arguments and result
!     type(four_momentum), dimension(:), intent(in) :: p
!     integer, dimension(:), intent(in) :: code
!     integer, intent(in) :: mode
!     logical :: accept
!     ! local variables
!     integer :: n, i, j
!     logical, dimension (size(code)) :: is_visible
!     ! number of particles
!     n = size (code)
!     ! flag the visible particles in a logical array
!     is_visible = visible (code)
!     ! There is one nontrivial set of cuts for mode=1
!     select case (mode)
!     case (1)
!        ! check if all particles (except invisible ones) are separated
!        accept = .true.
!        do i=1, n
!           if (is_visible(i)) then
!              do j=i+1, n
!                 if (is_visible(j) .and. eta_phi_distance (p(i), p(j)) < 0.4) &
!                      & accept = .false.
!              end do
!           end if
!        end do
!        if (.not.accept) return   ! No need to check further
!        ! Check for particles with minimum pT
!        accept = .false.
!        do i=1, n
!           ! Don't distinguish particle/antiparticle
!           select case (abs(code(i)))
!           case (E_LEPTON,MU_LEPTON,TAU_LEPTON)
!              if (transverse_momentum (p(i)) > 20)  accept = .true.
!           case (B_QUARK)
!              if (transverse_momentum (p(i)) > 80)  accept = .true.
!           case (GLUON, GLUON2, U_QUARK, D_QUARK, S_QUARK, C_QUARK)
!              if (transverse_momentum (p(i)) > 100)  accept = .true.
!           end select
!        end do
!        if (accept) return        ! No need to check further
!        ! Check the total visible and invisible tranverse momentum
!        if (transverse_momentum (sum (p, mask=is_visible)) > 50) then
!           accept = .true.
!           !     else if (transverse_momentum (sum (p, mask=.not.is_visible)) > 80) then
!           !        accept = .true.
!        end if
!     case default
!        ! No operation for other modes
!        accept = .true.
!     end select
!   end function cut
    

!========================================================================

! This weight function returns unity, i.e. no reweighting.
! Change this code if you need to reweight the matrix element according 
!   to a known (positive-semidefinite) function of the particle momenta
!   and activate it by setting user_weight_mode nonzero (the mode) in the
!   input file. 
! You may use the kinematical functions, operators and particle codes from the 
!   lorentz.f90 and particles.f90 modules for that purpose.

  function weight (p, code, mode)
    type(four_momentum), dimension(:), intent(in) :: p
    integer, dimension(:), intent(in) :: code
    integer, intent(in) :: mode
    real(kind=default) :: weight
    select case (mode)
    case default
       weight = 1
    end select
  end function weight
    
!========================================================================

! A standard interface for fragmentation has been defined in
!   E. Boos et al., Proc. Les Houches 2001, hep-ph/0109068
! The interface consists of two common blocks, HEPRUP and HEPEUP
! which are used to transfer all relevant information from the
! Monte Carlo generator to the fragmentation program.

! This subroutine can be modified for initializing user fragmentation.
! You can assume that the common block HEPRUP has been filled.
! Use HEPRUP to access all relevant information.

  subroutine fragment_init
    ! (insert code here)
  end subroutine fragment_init

! This subroutine can be modified for executing user fragmentation of
! a generated event.
! The fragmentation method is determined by the input variable
! fragmentation_method, unless it equals one of the built-in fragmentation
! methods (JETSET, PYTHIA).  It can be used to select among different versions
! of user fragmentation.
! You can assume that the common block HEPEUP has been filled.
! Use HEPEUP to access all relevant information.
! The result of fragmentation should be stored in the standard HEPEVT common
! block.

  subroutine fragment_call (method)
    integer, intent(in) :: method
    select case (method)
    case default
       ! (insert code here)
    end select
  end subroutine fragment_call

!=======================================================================

end module user
